import logging
from typing import Union

import numpy as np
import pandas as pd
import pegasus
from anndata import AnnData
from scipy.stats import norm
from statsmodels.stats.multitest import multipletests as padjust

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def compute_zscores(
    data: AnnData,
    layer_name: str,
    gene_symbols: str = None,
    cell_type_labels: str = "cellTypeOntologyID",
    verbose: bool = False,
) -> pd.DataFrame:
    """Calculate z-scores on an ``AnnData`` Object.

    Wrapper around `calculate_z_score` method from the pegasus package.

    Args:
        data:
            Input ``AnnData`` object.

        layer_name:
            Name of the layer containing the count matrix.
            Use None to use `X` slot of the AnnData object.

        gene_symbols:
            Column name from `var` that contains gene symbols.
            Defaults to `None`, to use var's index.

        celltype_annotation:
            Column name from `obs` that specifies cell type labels.
            Defaults to `None` for no cell type labels.

    Returns:
        A :py:class:`~pandas.DataFrame` of z-scores for each cell across all genes.
        This dataframe contains a column 'cell_types' that contains the cell type annotation labels.
    """
    logging.info("preprocessing data...")

    if gene_symbols is not None:
        if verbose:
            logging.info(f"setting column {gene_symbols} as index for `data.var`")
        data.var = data.var.set_index(gene_symbols, drop=False)
    else:
        logging.info("'var.index' is expected to contain gene symbols.")

    if verbose:
        logging.info("converting X to float32")

    if layer_name is None:
        data.X = data.X.astype(np.float32)
    else:
        data.X = data.layers[layer_name].astype(np.float32)

    if verbose:
        logging.info("calculating z-scores using pegasus")

    z_mat = pegasus.calculate_z_score(data)
    z_mat_df = pd.DataFrame(data=z_mat, columns=data.var.index.values)
    z_mat_df = z_mat_df.apply(pd.to_numeric, errors="coerce", downcast="float")

    if cell_type_labels is not None and cell_type_labels in data.obs.columns:
        if verbose:
            logging.info(f"using {cell_type_labels} as cell type annotation")
        z_mat_df["cell_types"] = data.obs[cell_type_labels].values
    else:
        if verbose:
            logging.info("Dataset does not contains cell type labels, using 'NA' as default.")
        z_mat_df["cell_types"] = "NA"

    z_mat_df.index = pd.Series(data.obs.index.values)
    # welp somethings gotta go because symbols are not unique
    z_mat_df = z_mat_df.loc[:, ~z_mat_df.columns.duplicated()]
    return z_mat_df


def compute_hist(
    x: Union[list, pd.Series],
    num_bins: int = 100,
    verbose: bool = False,
) -> dict:
    """Compute a histogram for a vector.

    Args:
        x:
            Input list or pandas Series object.

        num_bins:
            Number of bins.
            Defaults to 100.

    Returns:
        A dict containing the bins and the frequencies.
    """
    if verbose:
        logging.info("computing histogram bins")

    hist = np.histogram(x, bins=num_bins)
    res = {}
    res["bins"] = [f"{x:.2f}" for x in hist[1]]
    res["freq"] = hist[0]
    return res


def calculate_stats(
    zscores: Union[list, pd.Series],
    pvals: Union[list, pd.Series],
    num_hist_bins: int = 100,
    alpha: int = 0.1,
    verbose: bool = False,
) -> dict:
    """Calculate stats to for the dataset.

    Args:
        zscores:
            Mean zscore vector for genes of interest.

        pvals:
            Calculated pvalues.

        num_hist_bins:
            Number of histogram bins to generate. Defaults to 100.

        alpha:
            Alpha for error rate. Defaults to 0.1.

    Returns:
        A Dictionary containing the stats.
    """
    if verbose:
        logging.info("computing stats..")

    fdr_res = padjust(pvals, method="fdr_bh", alpha=alpha)
    mean_zscore = np.mean(zscores)
    set_zscore = mean_zscore * np.sqrt(len(zscores))
    set_pval = 1.0 - norm.cdf(set_zscore)
    zscoreGt2 = [z for z in zscores if z > 2]

    res = {
        "fdr_prop": len(zscoreGt2) / len(zscores),  # np.mean(fdr_res[0]) # % of  zscore > 2
        "total_count": len(fdr_res[0]),
        "fdr_count": len(zscoreGt2),  # np.sum(fdr_res[0]), # of  zscore > 2
        "zscore_hist": compute_hist(zscores, num_bins=num_hist_bins, verbose=verbose),
        "pval_hist": compute_hist(pvals, num_bins=num_hist_bins, verbose=verbose),
        "mean_zscore": mean_zscore,
        "set_zscore": set_zscore,
        "set_pval": set_pval,
    }

    return res



import logging
from typing import List, Union

import anndata
import pandas
from scipy.stats import norm

from .compute import calculate_stats, compute_zscores

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def signature_search(
    adata_or_path: Union[str, anndata.AnnData],
    genes: list,
    layer_name: str,
    gene_symbols_column_name: str = None,
    cell_type_label_column_name: str = None,
    num_hist_bins: int = 100,
    alpha: int = 0.1,
    include_celltype_enrichment: bool = False,
    verbose: bool = False,
) -> List[dict]:
    """Compute signature scores on a local dataset.

    Args:
        adata_or_path:
            Path to a H5ad file or :py:class:`~anndata.AnnData` object.

        genes:
            Input genes or features of interest.

        layer_name:
            Name of the layer containing the count matrix.

            Use None to use `X` slot of the AnnData object.

            This is expected to be consistent across all the anndata objects.

        gene_symbols_column_name:
            Column in the ``var`` slot of ``AnnData`` containing gene symbols.
            Defaults to the index of the ``var`` slot.

        cell_type_label_column_name:
            Column in the ``obs`` slof of ``AnnData`` containing cell type labels.
            Defaults to the index of the ``obs`` slot.

        num_hist_bins:
            Number of histrogram bins.
            Defaults to 100.

        alpha:
            Alpha for family wise error rate.
            Defaults to 0.1.

        include_celltype_enrichment:
            Wehether to group cells by celltype and compute enrichment scores.
            Must provide a ``celltype_label_column_name`` containing celltype labels.
            Defaults to False.

        verbose:
            Whether to display logs.
            Defaults to False.

    Returns:
        list containing results; each element in the list is a dictionary containng
        the cell type label, enrichment scores and other computed statistics.
    """
    results = []

    try:
        data = adata_or_path
        if isinstance(adata_or_path, str):
            data = anndata.read_h5ad(adata_or_path)

        z_mats = compute_zscores(
            data,
            gene_symbols=gene_symbols_column_name,
            cell_type_labels=cell_type_label_column_name,
            layer_name=layer_name,
            verbose=verbose,
        )

        missing_genes = list(set(genes).difference(set(z_mats.columns)))

        if len(missing_genes) > 0:
            msg = f"Input Dataset does not contain full set of input genes: {missing_genes}."
            if verbose:
                logging.error(msg)
            raise ValueError(msg)

        z_mat = z_mats[genes]
        zscores = z_mat.mean(axis=1)
        pvals = 1.0 - norm.cdf(zscores)
        res = calculate_stats(zscores, pvals, num_hist_bins=num_hist_bins, alpha=alpha)
        res["label"] = "overall"
        results.append(res)

        if include_celltype_enrichment:
            celltypes = pandas.DataFrame({"cell_types": z_mats["cell_types"], "zscores": zscores, "pvals": pvals})

            groups = celltypes.groupby("cell_types")

            if len(groups) > 1:
                for cname, group in groups:
                    res = calculate_stats(
                        group["zscores"].values,
                        group["pvals"].values,
                        num_hist_bins=num_hist_bins,
                        alpha=alpha,
                    )
                    res["label"] = cname
                    results.append(res)
    except Exception as e:
        logging.error(str(e))
        results = None
    finally:
        return results


def batch_signature_search(
    adatas_or_paths: List[Union[str, anndata.AnnData]],
    genes: List[str],
    gene_symbols_column_name: str = None,
    celltype_label_column_name: str = None,
    num_hist_bins: int = 100,
    alpha: float = 0.1,
    include_celltype_enrichment: bool = False,
    verbose: bool = False,
):
    """Compute signature scores on a collection of datasets (Batch mode).

    Args:
        adatas_or_paths:
            List of paths to a H5ad file or :py:class:`~anndata.AnnData` objects.

        genes:
            Input genes or features of interest.

        gene_symbols_column_name:
            Column in the ``var`` slot of ``AnnData`` containing gene symbols.
            Defaults to the index of the ``var`` slot.

        celltype_label_column_name:
            Column in the ``obs`` slof of ``AnnData`` containing cell type labels.
            Defaults to the index of the ``obs`` slot.

        num_hist_bins:
            Number of histrogram bins.
            Defaults to 100.

        alpha:
            Alpha for family wise error rate.
            Defaults to 0.1.

        include_celltype_enrichment:
            Wehether to group cells by celltype and compute enrichment scores.
            Must provide a ``celltype_label_column_name`` containing celltype labels.
            Defaults to False.

        verbose:
            Whether to display logs.
            Defaults to False.

    Returns:
        list containing results; each element in the list is a dictionary containng
        the cell type label, enrichment scores and other computed statistics.
    """
    results = []

    for path in adatas_or_paths:
        if verbose:
            logging.info(f"processing dataset: {path}")
        res = signature_search(
            path,
            genes=genes,
            gene_symbols_column_name=gene_symbols_column_name,
            celltype_label_column_name=celltype_label_column_name,
            num_hist_bins=num_hist_bins,
            alpha=alpha,
            include_celltype_enrichment=include_celltype_enrichment,
            verbose=verbose,
        )
        if res is not None:
            results.extend(res)

    return results



import os

from firmament.search import signature_search

__author__ = "kancherj"
__copyright__ = "kancherj"
__license__ = "MIT"


def test_sigsearch():
    """Test signature search on local H5AD."""
    results = signature_search(
        os.getcwd() + "/tests/data/test.h5ad",
        genes=["Tspan12", "Vipas39", "Cldn12"],
        layer_name=None,
        cell_type_label_column_name="level1class",
        include_celltype_enrichment=True,
        verbose=True,
    )

    assert isinstance(results, list)
    assert isinstance(results[0], dict)
